package Lingua::JA::Summarize;

use strict;
use warnings;
our $VERSION = 0.01;

use base qw(Class::Accessor::Fast Class::ErrorHandler);

use Carp;
use File::Temp qw(:POSIX);
use Jcode;


__PACKAGE__->mk_accessors(qw(mecab default_cost ng omit_number singlechar_factor));


sub NG () {
    my %map;
    @map{('(', ')', '#', ',')} = ();
    @map{qw(! " $ % & ' * + - . / : ; < = > ? @ [ \ ] ^ _ ` { | } ~)} = ();
    @map{
	qw(  ʬ    ǯ  ɥ
	        ϻ  Ȭ   ɴ    )} = ();
    @map{qw(     )} = ();
    \%map;
}

sub new {
    my ($proto, $fields) = @_;
    my $class = ref $proto || $proto;
    $fields = {} unless defined $fields;
    my $self = bless { %$fields }, $class;
    
    $self->{mecab} = 'mecab' unless $self->{mecab};
    $self->{default_cost} = 800 unless $self->{default_cost};
    $self->{ng} = NG unless defined $self->{ng};
    $self->{omit_number} = 1 unless defined $self->{omit_number};
    $self->{singlechar_factor} = 0.5 unless defined $self->{singlechar_factor};
    
    print STDERR "oh!\n" unless exists $self->{ng}->{'ǯ'};
    
    return $self;
}

sub keywords {
    my ($self, $args) = @_;
    $args = {} unless $args;
    my $threshold = $args->{threshold} || 5;
    my $maxwords = $args->{maxwords} || 5;
    my $stats = $self->{stats};
    my @keywords;
    
    foreach my $word (sort
		      { $stats->{$b}->{weight} <=> $stats->{$a}->{weight} ||
			    $a cmp $b }
		      keys(%$stats)) {
	last if $stats->{$word}->{weight} < $threshold;
	push(@keywords, $word);
	last if $maxwords == @keywords;
    }
    
    return @keywords;
}

sub analyze_file {
    my ($self, $file) = @_;
    
    my $fh;
    open($fh, '<', "$file") || croak("failed to open: $file: $!");
    my $slash = $/;
    undef $/;
    my $text = <$fh>;
    $/ = $slash;
    close $fh;
    
    $self->analyze($text);
}
    
sub analyze {
    my ($self, $text) = @_;
    my ($rfh, $wfh);
    
    croak("already analyzed") if $self->{stats};
    $self->{stats} = {};
    
    # adjust text
    $text .= "\n";
    $text =~ s/\s*\n\s*/\n/sg;
    $text = Jcode->new($text, 'euc')->tr('', '');
    $text =~ s//\n/g;
    # write text to temporary file
    my ($fh, $tempfile) = tmpnam();
    print $fh $text;
    close $fh;
    
    # open mecab
    my $mecab = $self->mecab;
    my $def_cost = $self->default_cost;
    open($fh, '-|',
	 $mecab .
	 " --node-format='%m\t%pn\t%pw\t%H\n'" .
	 " --unk-format='%m\t$def_cost\t$def_cost\tUnkType\n'" .
	 " --bos-format='\n'" .
	 " --eos-format='\n'" .
	 " $tempfile")
	|| croak("failed to call mecab ($mecab): $!");
    
    # read from mecab
    my $longword = {
	text => '',
	cost => 0,
	count => 0,
    };
    my $add_longword  = sub {
	if ($longword->{text}) {
	    $self->_add_word($longword->{text},
			     $longword->{cost} / $longword->{count});
	}
	$longword->{text} = '';
	$longword->{cost} = 0;
	$longword->{count} = 0;
    };
    while (my $line = <$fh>) {
	chomp($line);
	if ($line =~ /\t/o) {
	    my ($word, $pn, $pw, $H) = split(/\t/, $line, 4);
	    $word = $self->_normalize_word($word);
	    my $ng = $self->_ng_word($word);
	    if ($ng) {
		$add_longword->();
		next;
	    }
	    if ($H =~ /^̾/) {
		if ($H =~ /(Ω|̾)/) {
		    $add_longword->();
		    next;
		} elsif (! $longword->{text} && $H =~ //) {
		    # ng
		    next;
		}
	    } elsif ($H eq 'UnkType') {
		# handle unknown (mostly English) words
		if ($word =~ /^\w/ && $longword->{text} =~ /\w$/) {
		    $add_longword->();
		}
	    } else {
		$add_longword->();
		next;
	    }
	    $longword->{text} .= $word;
	    $longword->{cost} += $longword->{count} ? $pn : $pw;
	    $longword->{count}++;
	} else {
	    $add_longword->();
	}
    }
    $add_longword->();
    close $fh;
    unlink($tempfile);
    
    # calculate tf-idf
    $self->_calc_weight;
    
    1;
}
    
sub _add_word {
    my ($self, $word, $cost) = @_;
    return if $cost <= 0;
    my $target = $self->{stats}->{$word};
    if ($target) {
	$target->{count}++;
    } else {
	$self->{stats}->{$word} = { count => 1, cost => $cost };
    }
}

sub _calc_weight {
    my $self = shift;
    foreach my $word (keys(%{$self->{stats}})) {
	my $target = $self->{stats}->{$word};
	my $cost = $target->{cost};
	$cost = $self->default_cost unless $cost;
	$target->{weight} = $target->{count} * log(65536 / $cost);
	if (length($word) == 1) {
	    $target->{weight} *= $self->singlechar_factor;
	}
    }
}

sub _normalize_word {
    my ($self, $word) = @_;
    $word = Jcode->new($word, 'euc')->h2z;
    $word->tr('--ڣ-ʡ', '0-9A-Za-z()');
    lc($word);
}

sub _ng_word {
    my ($self, $word) = @_;
    return 1 if $self->omit_number && $word =~ /^\d*$/;
    return
	exists($self->{ng}->{$word}) ||
	exists($self->{ng}->{substr($word, 0, 1)});
}

1;
__END__

=head1 NAME

Lingua::JA::Summarize - A keyword extractor / summary generator

=head1 SYNOPSIS

    # Functional style
    
not yet
    
    # OO style
    
    use Lingua:::JA::Summarize;
    
    $s = Lingua::JA::Summarize->new;
    
    $s->analyze('Analyze some text');
    $s->analyze_file('filename_to_analyze.txt');

    my @keywords = $s->keywords;
    

=head1 DESCRIPTION

Lingua::JA::Summarize is a keyword extractor / summary generator for Japanese texts.  By using MeCab, the module extracts keywords from Japanese texts.

=head1 FUNCTIONS

to be written...

=head1 AUTHOR

Kazuho Oku E<lt>kazuhooku ___at___ gmail.comE<gt>

=head1 ACKNOWLEDGEMENTS

Thanks to Takesako-san for writing the prototype.

=head1 COPYRIGHT
Copyright (C) 2006  Cybozu Labs, Inc.

This library is free software; you can redistribute it and/or modify it under the same terms as Perl itself, either Perl version 5.8.7 or, at your option, any later version of Perl 5 you may have available.

=cut
